extern "C"
{
#include <wdm.h>
}
#include <dontuse.h>
#include <suppress.h>


typedef struct _MY_CONTEXT
{
	ULONG Data;
} MY_CONTEXT, *PMY_CONTEXT;

//	Forward Declarations
//
extern "C" DRIVER_INITIALIZE DriverEntry;
extern "C" NTSTATUS DriverEntry( __in PDRIVER_OBJECT pDriverObject, __in PUNICODE_STRING pRegistryPath );
NTSTATUS RfPreCreateOrOpenKeyEx( __in PVOID CallbackContext, __in PVOID Argument1, __in PREG_CREATE_KEY_INFORMATION CallbackData );
NTSTATUS RfPostCreateOrOpenKeyEx( __in PVOID CallbackContext, __in PVOID Argument1, __in PREG_POST_OPERATION_INFORMATION CallbackData );
NTSTATUS RfCallbackObjectContextCleanup( __in PVOID CallbackContext, __in PVOID Argument1, __in PREG_CALLBACK_CONTEXT_CLEANUP_INFORMATION CallbackData );
NTSTATUS RfRegistryCallback( __in PVOID CallbackContext, __in PVOID Argument1, __in PVOID Argument2 );
void RfUnload( __in PDRIVER_OBJECT pDriverObject );

//	Global Variables
//
LARGE_INTEGER g_CmCookie = { 0 };
PEX_CALLBACK_FUNCTION g_RegistryCallbackTable[ MaxRegNtNotifyClass ] = { 0 };
LONG g_nObjectContexts = 0;


extern "C"
NTSTATUS DriverEntry( __in PDRIVER_OBJECT pDriverObject, __in PUNICODE_STRING pRegistryPath )
{
    UNREFERENCED_PARAMETER( pRegistryPath );

    //	Set up our unload routine
    pDriverObject->DriverUnload = RfUnload;

    //	Set up our registry callback table
	g_RegistryCallbackTable[ RegNtPostCreateKeyEx ] = (PEX_CALLBACK_FUNCTION) RfPostCreateOrOpenKeyEx;
	g_RegistryCallbackTable[ RegNtPostOpenKeyEx ] = (PEX_CALLBACK_FUNCTION) RfPostCreateOrOpenKeyEx;
	g_RegistryCallbackTable[ RegNtPreCreateKeyEx ] = (PEX_CALLBACK_FUNCTION) RfPreCreateOrOpenKeyEx;
	g_RegistryCallbackTable[ RegNtPreOpenKeyEx ] = (PEX_CALLBACK_FUNCTION) RfPreCreateOrOpenKeyEx;
	g_RegistryCallbackTable[ RegNtCallbackObjectContextCleanup ] = (PEX_CALLBACK_FUNCTION) RfCallbackObjectContextCleanup;

    //	Register our callback with the system
    UNICODE_STRING AltitudeString = RTL_CONSTANT_STRING( L"360000" );
    NTSTATUS status = CmRegisterCallbackEx( RfRegistryCallback, &AltitudeString, pDriverObject, NULL, &g_CmCookie, NULL );
    if ( !NT_SUCCESS( status ) )
    {
        //	Failed to register - probably shouldn't succeed the driver intialization since this is critical
    }

    return status;
}

NTSTATUS RfPostCreateOrOpenKeyEx( __in PVOID CallbackContext, __in PVOID Argument1, __in PREG_POST_OPERATION_INFORMATION CallbackData )
{
	UNREFERENCED_PARAMETER( CallbackContext );
	UNREFERENCED_PARAMETER( Argument1 );
	PAGED_CODE();

	if ( CallbackData->CallContext )
	{
		if ( STATUS_SUCCESS == CallbackData->Status )
		{
			//	Attach the context we allocated to the opened object
			NTSTATUS status;
			status = CmSetCallbackObjectContext( CallbackData->Object, &g_CmCookie, CallbackData->CallContext, NULL );
			if ( NT_SUCCESS( status ) )
			{
				InterlockedIncrement( &g_nObjectContexts );
			}
			else
			{
				ExFreePool( CallbackData->CallContext );
			}
		}
		else
		{
			ExFreePool( CallbackData->CallContext );
		}
	}
	return STATUS_SUCCESS;
}

NTSTATUS RfPreCreateOrOpenKeyEx( __in PVOID CallbackContext, __in PVOID Argument1, __in PREG_CREATE_KEY_INFORMATION CallbackData )
{
	UNREFERENCED_PARAMETER( CallbackContext );
	UNREFERENCED_PARAMETER( Argument1 );
	PAGED_CODE();

	//	Allocate a new callback context and have it passed back to us when the open key completes
	PMY_CONTEXT pContext = (PMY_CONTEXT) ExAllocatePoolWithTag( PagedPool, sizeof(MY_CONTEXT), 'xCyM' );
	if ( pContext )
	{
		//	Put some meaningless data in our context that we can verify later
		pContext->Data = 42;
		CallbackData->CallContext = pContext;
	}
	return STATUS_SUCCESS;
}

NTSTATUS RfCallbackObjectContextCleanup( __in PVOID CallbackContext, __in PVOID Argument1, __in PREG_CALLBACK_CONTEXT_CLEANUP_INFORMATION CallbackData )
{
	UNREFERENCED_PARAMETER( CallbackContext );
	UNREFERENCED_PARAMETER( Argument1 );
	PAGED_CODE();

	if ( CallbackData->ObjectContext )
	{
		//	Make sure that our context looks like we expect it to
		PMY_CONTEXT pContext = (PMY_CONTEXT) CallbackData->ObjectContext;
		ASSERT( 42 == pContext->Data );
		ExFreePool( CallbackData->ObjectContext );
		InterlockedDecrement( &g_nObjectContexts );
	}
	return STATUS_SUCCESS;
}

NTSTATUS RfRegistryCallback( __in PVOID CallbackContext, __in PVOID Argument1, __in PVOID Argument2 )
{
	PAGED_CODE();
	
	REG_NOTIFY_CLASS Operation = (REG_NOTIFY_CLASS) (ULONG_PTR) Argument1;

    //	If we have no operation callback routine for this operation then just return to Cm
    if ( !g_RegistryCallbackTable[ Operation ] )
    {
        return STATUS_SUCCESS;
    }

    //	Call our operation callback routine
    return g_RegistryCallbackTable[ Operation ]( CallbackContext, Argument1, Argument2 );
}

void RfUnload( __in PDRIVER_OBJECT pDriverObject )
{
    UNREFERENCED_PARAMETER( pDriverObject );
    PAGED_CODE();

    NTSTATUS status = CmUnRegisterCallback( g_CmCookie );
    if ( !NT_SUCCESS( status ) )
    {
        //	Failed to unregister - try to handle gracefully
    }

	ASSERT( 0 == g_nObjectContexts );
}

